div.maplibregl-ctrl-bottom-left {
    width: 100%;
    margin-left: 0px;
    margin-right: 0px;
    padding-left: 0px;
    padding-right: 0px;
}

div.timeline-container {
    width: 100%;
    background-color: rgba(0, 0, 0, 0);
    margin-left: 0px;
    margin-right: 0px;
    padding-left: 1rem;
    padding-right: 1rem;
}

div#timeline {
    width: 100%;
    background-color: white;
    margin-left: 0px;
    margin-right: 0px;
    padding-left: 1rem;
    padding-right: 1rem;
    border-radius: 0.375rem;
    border: 1px solid rgba(0, 0, 0, 0.175);
}

#timeline-slider {
    width: 100%;
}

div.timeline-container datalist {
    display: flex;
    justify-content: space-between;
    width: 100%;
}
